----------------------------------------------------------------------------------------------------
-- Random Tour Lua Script
-- By Scepheo
----------------------------------------------------------------------------------------------------

----------------------------------------------------------------------------------------------------
-- Script configuration
----------------------------------------------------------------------------------------------------

-- How big the checkpoint should be in relation to the node
local nodeSizeMultiplier = 1.5

-- Distance after which a new checkpoint should be placed
local nodeMaxDistance = 50

-- Distance after which the first checkpoint should be placed
local nodeInitialDistance = 20

-- How many nodes (not checkpoints!) the path needs to contain
local minPathLength = 15

-- How sharp a corner needs to be (dot product) to place a waypoint after it
local angleThreshold = 0.9

-- Maximum distance between any two waypoints
local distanceThreshold = 150

-- Distance at which to draw debugging information
local debugDrawDistance = 100


----------------------------------------------------------------------------------------------------
-- Global variables
----------------------------------------------------------------------------------------------------

local graphpath = require('graphpath')
local settings = require('scenario/random_tour_settings')
math.randomseed(os.time())

local loaded = false
local running = false
local mapInstance = {}
local route


----------------------------------------------------------------------------------------------------
-- Helper functions
----------------------------------------------------------------------------------------------------

-- TODO: Replace with native BeamNG function
local function countItems(items)
	local itemCount = 0
	
	for key, value in pairs(items) do
		itemCount = itemCount + 1
	end
	
	return itemCount
end

-- TODO: Replace with native BeamNG function
local function getRandomValue(items)
	return items[tableChooseRandomKey(items)]
end

local waypointNumber = 1

local function addWaypointFor(scenario, routeEntry)
	local node = mapInstance.nodes[nodeName]
	
	local position = routeEntry.pos.x .. ' ' .. routeEntry.pos.y .. ' ' .. routeEntry.pos.z
	local wpName = 'random_tour_wp_' .. tostring(waypointNumber)
	waypointNumber = waypointNumber + 1

	TorqueScript.eval(
	[[
		if (isObject(]] .. wpName .. [[))
		{
			]] .. wpName .. [[.delete();
		}
		
		new BeamNGWaypoint(]] .. wpName .. [[)
		{
			drawDebug = "0";
			position = "]] .. position .. [[";
			rotation = "1 0 0 0";
			scale = "1 1 1";
			canSave = "1";
			canSaveDynamicFields = "1";
		};
		
		ScenarioObjectsGroup.add(]] .. wpName .. [[);
	]])
	
	table.insert(scenario.lapConfig, wpName)
	
	scenario.nodes[wpName] = 
	{
		pos = vec3(routeEntry.pos),
		radius = routeEntry.radius * nodeSizeMultiplier
	}
end

local function getDistanceBetween(aNode, bNode)
	if not aNode then return 0 end
	
	local a = mapInstance.nodes[aNode].pos
	local b = mapInstance.nodes[bNode].pos
	
	return (a - b):length()
end

local function getAngleBetween(aNode, bNode, cNode)
	if not aNode or not cNode then return 1 end
	
	local a = mapInstance.nodes[aNode].pos
	local b = mapInstance.nodes[bNode].pos
	local c = mapInstance.nodes[cNode].pos
	
	local x1 = b.x - a.x
	local y1 = b.y - a.y
	local l1 = math.sqrt(x1 * x1 + y1 * y1)
	x1 = x1 / l1
	y1 = y1 / l1
	
	local x2 = c.x - b.x
	local y2 = c.y - b.y
	local l2 = math.sqrt(x2 * x2 + y2 * y2)
	x2 = x2 / l2
	y2 = y2 / l2
	
	return x1 * x2 + y1 * y2
end


----------------------------------------------------------------------------------------------------
-- Time of Day
----------------------------------------------------------------------------------------------------

local duskStart  = 0.186
local nightStart = 0.229
local dawnStart  = 0.764
local dayStart   = 0.807

local function setVehicleLights()
	if not settings.setRandomTime then
		return
	end
	
	if scenetree.tod then
		local newTime = scenetree.tod.time
		local lightMode
		
		if newTime > nightStart and newTime < dawnStart then
			lightMode = '2'
		elseif newTime > duskStart and newTime < dayStart then
			lightMode = '1'
		else
			lightMode = '0'
		end
		
		scenetree.scenario_player0:queueLuaCommand('electrics.setLightsState('..lightMode..')')
	end
end

local function setRandomTime()
	if not settings.setRandomTime then
		return
	end

	if scenetree.tod then
		-- select dawn, day, dusk or night
		local partOfDay = math.random(1, 4)
		local newTime
		
		if partOfDay == 1 then
			log('I', 'random_tour', 'setRandomTime: Selected dawn')
			newTime = dawnStart + math.random() * (dayStart - dawnStart)
		elseif partOfDay == 2 then
			log('I', 'random_tour', 'setRandomTime: Selected day')
			newTime = (dayStart + math.random() * (1 + duskStart - dayStart)) % 1
		elseif partOfDay == 2 then
			log('I', 'random_tour', 'setRandomTime: Selected dusk')
			newTime = duskStart + math.random() * (nightStart - duskStart)
		else
			log('I', 'random_tour', 'setRandomTime: Selected night')
			newTime = nightStart + math.random() * (dawnStart - nightStart)
		end
		
		log('I', 'random_tour', 'setRandomTime: Selected time '..tostring(newTime))	
		scenetree.tod.time = newTime
	end
end


----------------------------------------------------------------------------------------------------
-- Random vehicle selection
----------------------------------------------------------------------------------------------------

local function loadPlayer(position, direction)
	
	-- Set the player at the starting location
	local tsPosition = position.x .. ' ' .. position.y .. ' ' .. position.z
	
	local angle = math.atan2(direction.x, direction.y)
	angle = math.deg(angle) + 180
	local tsRotation = '0 0 1 ' .. angle
	
	local configList = vehicles.getConfigList().configs
	
	local vehicle
	local valid = false
	
	while not valid do
		vehicle = getRandomValue(configList)
		
		for _, validType in pairs(settings.allowedVehicleTypes) do
			if vehicle.aggregates.Type[validType] then
				valid = true
			end
		end
	end
	
	local jbeam = vehicle.model_key
	local configDisplay = vehicle.Configuration or '(Default configuration)'
	
	log('I', 'random_tour.loadPlayer',
		string.format('Selected %s %s', jbeam, configDisplay))
	
	local configuration = string.format('vehicles/%s/%s.pc', jbeam, vehicle.key)
	
	local color = vehicle.default_color or '0 0 0 1.25'
	
	TorqueScript.eval(
	[[
		scenario_player0.setTransform(matrixCreate("]] .. tsPosition .. [[", "]] .. tsRotation .. [["));
		scenario_player0.JBeam = "]] .. jbeam .. [[";
		scenario_player0.partConfig = "]] .. configuration .. [[";
		scenario_player0.color = "]] .. color .. [[";
		scenario_player0.requestReload();
	]])
end


----------------------------------------------------------------------------------------------------
-- Random route generation
----------------------------------------------------------------------------------------------------

local function getRandomPath()

	-- Build gp
	local gp = graphpath.newGraphpath()
	
    for k, n in pairs(mapInstance.nodes) do
        n.pos = vec3(n.pos)
        gp:setPointPosition(k, n.pos)
        n.drivability = 1
    end
	
    for nid, n in pairs(mapInstance.nodes) do
        for lid, dif in pairs(n.links) do
            local totaldif = mapInstance.nodes[nid].pos:distance(mapInstance.nodes[lid].pos) * (1 / (dif + 1e-30))
            gp:bidiEdge(nid, lid, totaldif)
        end
    end
	
	-- Select a path that's possible and longer than the minimum length
	local path = nil
	
	while path == nil or #path < minPathLength do
		startNodeName = tableChooseRandomKey(mapInstance.nodes)
		endNodeName = startNodeName
		
		while endNodeName == startNodeName do
			endNodeName = tableChooseRandomKey(mapInstance.nodes)
		end
		
		path = gp:getPath(startNodeName, endNodeName)
		
		local thirdNode = tableChooseRandomKey(mapInstance.nodes)
		local otherPath = gp:getPath(endNodeName, thirdNode)
		local tries = 5
		
		while tries > 0 and (otherPath == nil or #otherPath < 2 or otherPath[2] == path[#path - 1]) do
			thirdNode = tableChooseRandomKey(mapInstance.nodes)
			otherPath = gp:getPath(endNodeName, thirdNode)
			tries = tries - 1
		end
		
		if tries == 0 then
			path = nil
		else
			for i = 2, #otherPath do
				table.insert(path, otherPath[i])
			end
		end
	end
	
	return path
end

local function makeRoute(path)
	local result = {}
	local distanceToStart = 0
	
	for i = 1, #path do
		local node = mapInstance.nodes[path[i]]
		local distanceToLast = getDistanceBetween(path[i - 1], path[i])
		
		table.insert(result, {
			name = path[i],
			pos = node.pos,
			radius = node.radius,
			road = string.match(path[i], "(.-)%d+$"),
			angle = getAngleBetween(path[i - 1], path[i], path[i + 1]),
			distanceToLast = distanceToLast,
			distanceToStart = distanceToStart,
			linkCount = countItems(node.links)
		})
		
		distanceToStart = distanceToStart + distanceToLast
	end
	
	return result
end

local function indexOfInterest(index, recurse)
	recurse = (recurse == nil) or recurse
	
	if route[index].angle < angleThreshold then
		return true
	end
	
	if route[index].linkCount >= 2 and route[index].angle < 0.95 then
		return true
	end
	
	if recurse and index < #path then
		return indexOfInterest(index + 1, false)
	end
	
	return false
end


----------------------------------------------------------------------------------------------------
-- Scenario events
----------------------------------------------------------------------------------------------------

local function onLoad(scenario)
	setRandomTime()
	
	mapInstance = map.getMap()
	
	local path = getRandomPath()
	route = makeRoute(path)
	
	log('I', 'random_tour.onLoad', 'Got random path:')
	
	for i, nodeName in ipairs(path) do
		log('I', 'random_tour.onLoad', string.format('%003d. %s', i, nodeName))
	end
	
	scenario.lapConfig = {}
	
	local startPosition = mapInstance.nodes[path[1]].pos
	local previousPosition = startPosition
	local startDirection = nil
	local currentDistance = nodeMaxDistance - nodeInitialDistance
	
	local waypointIndices = {}
	local waypoints = {}
	
	-- Create first waypoint
	local i = 1
	while route[i].distanceToStart < nodeInitialDistance do
		i = i + 1
	end
	
	table.insert(waypointIndices, i)
	table.insert(waypoints, route[i])
	
	-- Create waypoints for corner nodes
	while i < #path - 1 do
		if indexOfInterest(i) then
			while indexOfInterest(i) do
				i = i + 1
			end
			
			local startNode = route[i - 2]
			local endNode = route[i - 1]
			
			local c = startNode.radius * 3
			local dir = endNode.pos - startNode.pos
			local length = dir:length()
			local offset = Point3F(dir.x * c / length, dir.y * c / length, dir.z * c / length)
			route[i - 1].pos = startNode.pos + offset
			
			table.insert(waypointIndices, i - 1)
			table.insert(waypoints, route[i])
		else
			i = i + 1
		end
	end
	
	-- Insert last waypoint
	table.insert(waypointIndices, #route)
	table.insert(waypoints, route[#route])
	
	-- Fill in large gaps between nodes
	local insertIndex = 2
	for i = 2, #waypointIndices do
		local firstIndex = waypointIndices[i - 1]
		local firstNode = route[firstIndex]
		local secondIndex = waypointIndices[i]
		local secondNode = route[secondIndex]
		log('D', 'random_tour', 'Checking ' .. tostring(firstIndex) .. ' (' .. firstNode.name .. ') - '
			.. tostring(secondIndex) .. ' (' .. secondNode.name .. ')')
		
		local distance = secondNode.distanceToStart - firstNode.distanceToStart
		log('D', 'random_tour', 'Distance ' .. tostring(distance))
		
		if distance > distanceThreshold then
			local count = math.floor(distance / distanceThreshold)
			local stepSize = distance / (count + 1)
			
			log('D', 'random_tour', 'Inserting extra waypoints for stepSize ' .. tostring(stepSize))
			
			local currentDistance = 0
			
			for j = firstIndex + 1, secondIndex - 1 do
				local step = route[j].distanceToStart - route[j - 1].distanceToStart
				currentDistance = currentDistance + step
				
				if currentDistance >= stepSize then
					currentDistance = currentDistance - stepSize
					
					log('D', 'random_tour', 'Inserted ' .. route[j].name .. ' at index ' .. tostring(insertIndex))
					table.insert(waypoints, insertIndex, route[j])
					insertIndex = insertIndex + 1
				end
			end
		end
		
		insertIndex = insertIndex + 1
	end
	
	log('D', 'random_tour', 'Route:')
	for i = 1, #waypoints do
		local text
		if waypoints[i] then
			text = waypoints[i].name
		else
			text = 'nil'
		end
		log('D', 'random_tour', text)
	end
	
	if startDirection == nil then
		startDirection = waypoints[1].pos - startPosition
	end
	
	for i = 1, #waypoints do
		addWaypointFor(scenario, waypoints[i])
	end
	
	loadPlayer(startPosition, startDirection)
end

local function onScenarioChange(scenario)
	if scenario and scenario.state == 'pre-start' then
		log('I', 'random_tour.onScenarioChange', 'Scenario started, starting set-up')
		onLoad(scenario)
		log('I', 'random_tour.onScenarioChange', 'Scenario set-up complete')
	end
end

local function onPhysicsEngineEvent(eventType)
	-- Freeze whenever a vehicle is spawned
	if eventType == 'spawn' then
		log('I', 'random_tour.onPhysicsEngineEvent', 'Vehicle spawned, freezing scenario_player0')
		scenetree.scenario_player0:queueLuaCommand('scenario.freeze(1)')
		setVehicleLights()
	end
end


----------------------------------------------------------------------------------------------------
-- Debugging
----------------------------------------------------------------------------------------------------

require('mathlib')
local textColor = ColorF(0, 0, 0, 0.5)
local straightLineColor = ColorF(0, 1, 0, 0.8)
local angledLineColor = ColorF(1, 0, 0, 0.8)

local function addZ(position, z)
	return vec3(position.x, position.y, position.z + z)
end

local function debugText(routeEntry)
	return String(string.format("%d (%.2f)", routeEntry.linkCount, routeEntry.angle))
end

local function onDrawDebug(focusPosition)
	if not route then return end
	
	local firstPos = addZ(route[1].pos, 1)
	local firstPosDistance = (firstPos - focusPosition):length()
	
	if firstPosDistance < debugDrawDistance then
		debugDrawer:drawText(firstPos:toPoint3F(), debugText(route[1]), textColor)
	end
	
	for i = 2, #route do
		local secondPos = addZ(route[i].pos, 1)
		local secondPosDistance = (secondPos - focusPosition):length()
		
		if firstPosDistance < debugDrawDistance or secondPosDistance < debugDrawDistance then
			local color
			
			if route[i - 1].angle < angleThreshold then
				color = angledLineColor
			else
				color = straightLineColor
			end
			
			debugDrawer:drawCylinder(firstPos:toPoint3F(), secondPos:toPoint3F(), 0.04, color)
		end
		
		if secondPosDistance < debugDrawDistance then
			debugDrawer:drawText(secondPos:toPoint3F(), debugText(route[i]), textColor)
		end
		
		firstPos = secondPos
		firstPosDistance = secondPosDistance
	end
end


----------------------------------------------------------------------------------------------------
-- Module interface
----------------------------------------------------------------------------------------------------

return {
	onPhysicsEngineEvent = onPhysicsEngineEvent,
	onScenarioChange = onScenarioChange
	--onDrawDebug = onDrawDebug
}